#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Deformation functions for Amen.
All functions act on an Audio object and return a modified Audio object.
"""

import librosa

from .audio import Audio


def time_stretch(audio, rate):
    """
    Wraps librosa's `time_stretch` function, and returns a new Audio object.
    Note that this folds to mono.

    Parameters
    ---------
    audio : Audio
        The Audio object to act on.

    rate : float
        The stretch factor.
        If rate is > 1, then the audio is sped up.
        If rate is < 1, then the audio is slowed down.
        A rate of `2.0` will result in audio that is twice as fast.
    """
    stretched = librosa.effects.time_stretch(
        librosa.to_mono(audio.raw_samples), rate=rate
    )
    stretched_audio = Audio(raw_samples=stretched, sample_rate=audio.sample_rate)

    return stretched_audio


def pitch_shift(audio, steps, step_size=12):
    """
    Wraps librosa's `pitch_shift` function, and returns a new Audio object.
    Note that this folds to mono.

    Parameters
    ---------
    audio : Audio
        The Audio object to act on.

    steps : float
        The pitch shift amount.
        The default unit is semitones, as set by `step_size`.

    step_size : float > 0
        The number of equal-tempered steps per octave.
        The default is semitones, as set by `step_size=12`.
        Quarter-tones, for example, would be `step_size=24`.
    """
    shifted = librosa.effects.pitch_shift(
        librosa.to_mono(audio.raw_samples),
        audio.sample_rate,
        steps,
        bins_per_octave=step_size,
    )
    stretched_audio = Audio(raw_samples=shifted, sample_rate=audio.sample_rate)

    return stretched_audio


def harmonic_separation(audio, margin=3.0):
    """
    Wraps librosa's `harmonic` function, and returns a new Audio object.
    Note that this folds to mono.

    Parameters
    ---------
    audio : Audio
        The Audio object to act on.

    margin : float
        The larger the margin, the larger the separation.
        The default is `3.0`.
    """
    harmonic = librosa.effects.harmonic(
        librosa.to_mono(audio.raw_samples), margin=margin
    )
    harmonic_audio = Audio(raw_samples=harmonic, sample_rate=audio.sample_rate)

    return harmonic_audio


def percussive_separation(audio, margin=3.0):
    """
    Wraps librosa's `percussive` function, and returns a new Audio object.
    Note that this folds to mono.

    Parameters
    ---------
    audio : Audio
        The Audio object to act on.

    margin : float
        The larger the margin, the larger the separation.
        The default is `3.0`.
    """
    percussive = librosa.effects.percussive(
        librosa.to_mono(audio.raw_samples), margin=margin
    )
    percussive_audio = Audio(raw_samples=percussive, sample_rate=audio.sample_rate)

    return percussive_audio



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Audio analysis
'''

import os
import pandas as pd
import numpy as np
import soundfile as sf

import librosa

from .feature import Feature, FeatureCollection
from .timing import TimingList


class Audio(object):
    """
    The base Audio object:  wraps the ouput from librosa, and provides access to features

    Attributes
    ----------
        sample_rate: number
            sample rate
        raw_samples: numpy array
            raw samples from the audio
        analysis_samples: numpy array
            downsampled samples for analysis
        num_channels: integer
            number of channels of the audio
        duration: float
            duration, in seconds
        features: dict
            collection of named feature objects
    """

    def __init__(
        self,
        file_path=None,
        raw_samples=None,
        convert_to_mono=False,
        sample_rate=44100,
        analysis_sample_rate=22050,
    ):
        """
        Audio constructor.
        Opens a file path, loads the audio with librosa, and prepares the features

        Parameters
        ----------

        file_path: string
            path to the audio file to load

        raw_samples: np.array
            samples to use for audio output

        convert_to_mono: boolean
            (optional) converts the file to mono on loading

        sample_rate: number > 0 [scalar]
            (optional) sample rate to pass to librosa.


        Returns
        ------
        An Audio object
        """

        if file_path:
            y, sr = librosa.load(file_path, mono=convert_to_mono, sr=sample_rate)
        elif raw_samples is not None:
            # This assumes that we're passing in raw_samples
            # directly from another Audio's raw_samples.
            y = raw_samples
            sr = sample_rate

        self.file_path = file_path
        self.sample_rate = float(sr)
        self.analysis_sample_rate = float(analysis_sample_rate)
        self.num_channels = y.ndim
        self.duration = librosa.get_duration(y=y, sr=sr)

        self.analysis_samples = librosa.resample(
            librosa.to_mono(y), sr, self.analysis_sample_rate, res_type='kaiser_best'
        )
        self.raw_samples = np.atleast_2d(y)

        self.zero_indexes = self._create_zero_indexes()
        self.features = self._create_features()
        self.timings = self._create_timings()

    def __repr__(self):
        file_name = os.path.split(self.file_path)[-1]
        args = file_name, self.duration
        return '<Audio, file: {0:s}, duration: {1:.2f}>'.format(*args)

    def output(self, filename, format=None):
        """
        Write the samples out to the given filename.

        Parameters
        ----------
        filename : str
            The path to write the audio on disk.
            This can be any format supported by `pysoundfile`, including
            `WAV`, `FLAC`, or `OGG` (but not `mp3`).

        format : str
            If provided, explicitly set the output encoding format.
            See `soundfile.available_formats`.
        """
        sf.write(filename, self.raw_samples.T, int(self.sample_rate), format=format)

    def _create_zero_indexes(self):
        """
        Create zero crossing indexes.
        We use these in synthesis, and it is easier to make them here.
        """
        zero_indexes = []
        for channel_index in range(self.num_channels):
            channel = self.raw_samples[channel_index]
            zero_crossings = librosa.zero_crossings(channel)
            zero_index = np.nonzero(zero_crossings)[0]
            zero_indexes.append(zero_index)
        return zero_indexes

    def _create_timings(self):
        """
        Create timings in a timings dict.
        """
        timings = {}
        timings['track'] = TimingList('track', [(0, self.duration)], self)
        timings['beats'] = TimingList('beats', self._get_beats(), self)
        timings['segments'] = TimingList('segments', self._get_segments(), self)
        return timings

    def _get_beats(self):
        """
        Gets beats using librosa's beat tracker.
        """
        _, beat_frames = librosa.beat.beat_track(
            y=self.analysis_samples, sr=self.analysis_sample_rate, trim=False
        )

        # pad beat times to full duration
        f_max = librosa.time_to_frames(self.duration, sr=self.analysis_sample_rate)
        beat_frames = librosa.util.fix_frames(beat_frames, x_min=0, x_max=f_max)

        # convert frames to times
        beat_times = librosa.frames_to_time(beat_frames, sr=self.analysis_sample_rate)

        # make the list of (start, duration) tuples that TimingList expects
        starts_durs = [(s, t - s) for (s, t) in zip(beat_times, beat_times[1:])]

        return starts_durs

    def _get_segments(self):
        """
        Gets Echo Nest style segments using librosa's onset detection and backtracking.
        """

        onset_frames = librosa.onset.onset_detect(
            y=self.analysis_samples, sr=self.analysis_sample_rate, backtrack=True
        )
        segment_times = librosa.frames_to_time(
            onset_frames, sr=self.analysis_sample_rate
        )

        # make the list of (start, duration) tuples that TimingList expects
        starts_durs = [(s, t - s) for (s, t) in zip(segment_times, segment_times[1:])]

        return starts_durs

    def _create_features(self):
        """
        Creates the FeatureCollection, and loads each feature.

        Parameters
        ---------

        Returns
        -----
        FeatureCollection
            FeatureCollection with each Amen.Feature object named correctly.
            Note that _get_chroma returns a FeatureCollection of chroma features.
        """
        features = FeatureCollection()
        features['centroid'] = self._get_centroid()
        features['amplitude'] = self._get_amplitude()
        features['timbre'] = self._get_timbre()
        features['chroma'] = self._get_chroma()
        features['tempo'] = self._get_tempo()
        return features

    def _get_centroid(self):
        """
        Gets spectral centroid data from librosa, and returns it as a Feature

        Parameters
        ---------

        Returns
        -----
        Feature
        """
        centroids = librosa.feature.spectral_centroid(self.analysis_samples)
        data = self._convert_to_dataframe(centroids, ['spectral_centroid'])
        feature = Feature(data)
        return feature

    def _get_amplitude(self):
        """
        Gets amplitude data from librosa, and returns it as a Feature

        Parameters
        ---------

        Returns
        -----
        Feature
        """
        amplitudes = librosa.feature.rms(self.analysis_samples)
        data = self._convert_to_dataframe(amplitudes, ['amplitude'])
        feature = Feature(data)
        return feature

    def _get_timbre(self):
        """
        Gets timbre (MFCC) data, taking the first 20.
        Note that the keys to the Feature are "mffc_<index>",
        to avoid having a dict-like object with numeric keys.

        Parameters
        ---------

        Returns
        -----
        Feature
        """
        mfccs = librosa.feature.mfcc(
            y=self.analysis_samples, sr=self.analysis_sample_rate, n_mfcc=12
        )
        feature = FeatureCollection()
        for index, mfcc in enumerate(mfccs):
            data = self._convert_to_dataframe(mfcc, ['timbre'])
            key = 'mfcc_%s' % (index)
            feature[key] = Feature(data)

        return feature

    def _get_chroma(self):
        """
        Gets chroma data from librosa, and returns it as a FeatureCollection,
        with 12 features.

        Parameters
        ---------

        Returns
        -----
        FeatureCollection
        """
        feature = FeatureCollection()
        pitch_names = ['c', 'c#', 'd', 'eb', 'e', 'f', 'f#', 'g', 'ab', 'a', 'bb', 'b']
        chroma_cq = librosa.feature.chroma_cqt(self.analysis_samples)
        for chroma, pitch in zip(chroma_cq, pitch_names):
            data = self._convert_to_dataframe(chroma, [pitch])
            feature[pitch] = Feature(data)

        # Enharmonic aliases
        feature['db'] = feature['c#']
        feature['d#'] = feature['eb']
        feature['gb'] = feature['f#']
        feature['g#'] = feature['ab']
        feature['a#'] = feature['bb']

        return feature

    def _get_tempo(self):
        """
        Gets tempo data from librosa, and returns it as a feature collection.
        Note that the tempo feature uses median aggregation, as opposed to the
        default mean.

        Parameters
        ---------

        Returns
        -----
        FeatureCollection
        """
        onset_env = librosa.onset.onset_strength(
            self.analysis_samples, sr=self.analysis_sample_rate
        )
        tempo = librosa.beat.tempo(
            onset_envelope=onset_env, sr=self.analysis_sample_rate, aggregate=None
        )
        data = self._convert_to_dataframe(tempo, ['tempo'])
        feature = Feature(data, aggregate=np.median)

        return feature

    @classmethod
    def _convert_to_dataframe(cls, feature_data, columns):
        """
        Take raw librosa feature data, convert to a pandas dataframe.

        Parameters
        ---------
        feature_data: numpy array
            a N by T array, where N is the number of features, and T is the number of time dimensions

        columns: list [strings]
            a list of column names of length N, the same as the N dimension of feature_data

        Returns
        -----
        pandas.DataFrame
        """
        feature_data = feature_data.transpose()
        frame_numbers = np.arange(len(feature_data))
        indexes = librosa.frames_to_time(frame_numbers)
        indexes = pd.to_timedelta(indexes, unit='s')
        data = pd.DataFrame(data=feature_data, index=indexes, columns=columns)
        return data



#!/usr/bin/env python
'''Timing interface'''

from bisect import bisect_left, bisect_right

import numpy as np
import pandas as pd

import librosa


class TimeSlice(object):
    """
    A slice of time:  has a start time, a duration, and a reference to an Audio object.
    """

    def __init__(self, time, duration, audio, unit='s'):
        self.time = pd.to_timedelta(time, unit=unit)
        self.duration = pd.to_timedelta(duration, unit=unit)
        self.audio = audio

    def __repr__(self):
        args = self.time.delta * 1e-9, self.duration.delta * 1e-9
        return '<TimeSlice, start: {0:.2f}, duration: {1:.2f}>'.format(*args)

    def get_samples(self):
        """
        Gets the samples corresponding to this TimeSlice from the parent audio object.
        """
        start = self.time.delta * 1e-9
        duration = self.duration.delta * 1e-9
        starting_sample, ending_sample = librosa.time_to_samples(
            [start, start + duration], self.audio.sample_rate
        )

        left_offsets, right_offsets = self._get_offsets(
            starting_sample, ending_sample, self.audio.num_channels
        )

        samples = self._offset_samples(
            starting_sample,
            ending_sample,
            left_offsets,
            right_offsets,
            self.audio.num_channels,
        )

        return samples, left_offsets[0], right_offsets[0]

    def _get_offsets(self, starting_sample, ending_sample, num_channels):
        """
        Find the offset to the next zero-crossing, for each channel.
        """
        offsets = []
        for zero_index in self.audio.zero_indexes:
            index = bisect_left(zero_index, starting_sample) - 1
            if index < 0:
                starting_offset = 0
            else:
                starting_crossing = zero_index[index]
                starting_offset = starting_crossing - starting_sample

            index = bisect_left(zero_index, ending_sample)
            if index >= len(zero_index):
                ending_offset = 0
            else:
                zci = min(bisect_right(zero_index, ending_sample), len(zero_index) - 1)
                ending_crossing = zero_index[zci]
                ending_offset = ending_crossing - ending_sample

            offsets.append((starting_offset, ending_offset))

        if num_channels == 1:
            results = (offsets[0], offsets[0])
        elif num_channels == 2:
            results = (offsets[0], offsets[1])

        return results

    def _offset_samples(
        self, starting_sample, ending_sample, left_offsets, right_offsets, num_channels
    ):
        """
        Does the offset itself.
        """
        left_slice = (
            0,
            slice(starting_sample + left_offsets[0], ending_sample + left_offsets[1]),
        )
        right_slice = left_slice

        if num_channels == 2:
            right_slice = (
                1,
                slice(
                    starting_sample + right_offsets[0], ending_sample + right_offsets[1]
                ),
            )

        left_channel = self.audio.raw_samples[left_slice]
        right_channel = self.audio.raw_samples[right_slice]
        return np.array([left_channel, right_channel])


class TimingList(list):
    """
    A list of TimeSlices.
    """

    def __init__(self, name, timings, audio, unit='s'):

        super(self.__class__, self).__init__()

        self.name = name
        for (start, duration) in timings:
            time_slice = TimeSlice(start, duration, audio, unit=unit)
            self.append(time_slice)



#!/usr/bin/env python
'''Container classes for feature analysis'''

import numpy as np
import pandas as pd
import six

from .timing import TimeSlice
from .exceptions import FeatureError


class Feature(object):
    """
    Core feature container object.  Handles indexing and time-slicing.

    Attributes
    ---------

    Methods
    ------
    at(time_slices)
        Resample the feature at the given TimeSlices
    """

    def __init__(self, data, aggregate=np.mean, base=None, time_slices=None):
        """
        Constructor for feature object

        Parameters
        ----------
        data: pandas.DataFrame
            Time-indexed data frame of features

        aggregate: function
            resample-aggregation function or mapping

        Returns
        ------
        A Feature object
        """

        # Check that the arguments have the right types
        assert isinstance(data, pd.DataFrame)

        self.data = data
        self.aggregate = aggregate
        self.time_slices = time_slices
        # Not sure that this is the right way to do it - I feel like we're outsmarting pandas
        # pandas supports multiple keys in a dataframe, whereas this only allows one.
        # Should we replace FeatureCollection with something like that?
        self.name = data.keys()[0]

        if base is not None:
            assert isinstance(base, Feature)

        self.base = base

    def __repr__(self):
        args = self.name
        return '<Feature, {0}>'.format(args)

    def __iter__(self):
        """
        Wrapper to allow easy access to the internal data of the pandas dataframe
        """
        for datum in self.data[self.name]:
            yield datum

    def __getitem__(self, x):
        """
        Wrapper to allow easy access to the internal data of the pandas dataframe
        """
        return self.data[self.name][x]

    def __len__(self):
        """
        Wrapper to allow easy access to the internal data of the pandas dataframe
        """
        return len(self.data[self.name])

    def with_time(self):
        """
        Allows iteration over a time-indexed feature and the associated timeslices.
        """
        if self.time_slices is None:
            raise FeatureError("Feature has no time reference.")

        for i, datum in enumerate(self.data[self.name]):
            yield (self.time_slices[i], datum)

    def at(self, time_slices):
        """
        Resample the data at a new time slice index.

        Parameters
        ----------
        time_slices: TimeSlice or TimeSlice collection
            The time slices at which to index this feature object

        Returns
        -------
        Feature
            The resampled feature data
        """

        if self.base is not None:
            return self.base.at(time_slices)

        if isinstance(time_slices, TimeSlice):
            time_slices = [time_slices]

        # join the time slice values
        timed_data = pd.DataFrame(columns=self.data.columns)

        # make the new data
        for slice_t in time_slices:
            slice_index = (slice_t.time <= self.data.index) & (
                self.data.index < slice_t.time + slice_t.duration
            )
            timed_data.loc[slice_t.time] = self.aggregate(
                self.data[slice_index], axis=0
            )

        # return the new feature object
        return Feature(
            data=timed_data,
            aggregate=self.aggregate,
            base=self,
            time_slices=time_slices,
        )


class FeatureCollection(dict):
    """
    A dictionary of features.

    Delegates `.at` to the features it contains.

    Allows for selection of multiple keys, which returns a smaller feature collection.
    """

    def at(self, time_slices):
        """
        Resample each feature at a new time slice index.

        Parameters
        ----------
        time_slices : TimeSlice or TimeSlice collection
            The time slices at which to index this feature object

        Returns
        -------
        new_features : FeatureCollection
            The resampled feature data
        """
        new_features = FeatureCollection()
        for key in self.keys():
            new_features[key] = self[key].at(time_slices)
        return new_features

    def __iter__(self):
        """
        Wrapper to avoid making the user deal with parallel lists
        """
        key = list(self.keys())[0]
        length = len(self[key])
        for i in range(length):
            res = {}
            for key, feature in self.items():
                res[key] = feature.data[feature.name][i]
            yield res

    def __len__(self):
        """
        Wrapper to avoid making the user deal with parallel lists
        """
        key = list(self.keys())[0]
        feature = self[key]
        return len(feature)

    def with_time(self):
        """
        Allows iteration over a time-indexed feature and the associated timeslices.
        """
        key = list(self.keys())[0]
        length = len(self[key])
        time_slices = self[key].time_slices

        if time_slices is None:
            raise FeatureError("FeatureCollection has no time reference.")

        for i in range(length):
            res = {}
            for key, feature in self.items():
                res[key] = feature.data[feature.name][i]
            yield (time_slices[i], res)

    def get(self, keys):
        """
        Get a subset of the keys in the correct feature collection

        Parameters
        ----------
        keys : A string or list of strings
            The keys to return from the current feature collection

        Returns
        -------
        new_features : FeatureCollection
            The subset of keys
        """
        if isinstance(keys, six.string_types):
            keys = [keys]

        new_features = FeatureCollection()
        for key in keys:
            if key in self:
                new_features[key] = self[key]
        return new_features



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Audio synthesis'''

import types
import pandas as pd
from scipy.sparse import lil_matrix

import librosa

from .audio import Audio
from .exceptions import SynthesizeError


def _format_inputs(inputs):
    """
    Organizes inputs to be a list of (TimeSlice, start_time) tuples,
    if they are not already of that form.
    # We may want to update this to support "properly" zipped lists of tuples.

    Acceptable forms are:
        - A list of TimeSlices - the TimeSlices will be concatenated.
        - A generator that returns (TimeSlice, start_time).
        - A tuple of (TimeSlices, start_times).
    """
    if isinstance(inputs, list):
        time_index = pd.to_timedelta(0.0, 's')
        timings = []
        for time_slice in inputs:
            timings.append(time_index)
            time_index = time_index + time_slice.duration

        return zip(inputs, timings)

    elif isinstance(inputs, tuple):
        return zip(inputs[0], inputs[1])

    elif isinstance(inputs, types.GeneratorType):
        return inputs
    raise SynthesizeError('Invalid synthesis timing format: {}'.format(inputs))


def synthesize(inputs):
    """
    Generate new Audio objects for output or further remixing.

    Parameters
    ----------

    inputs: generator, list, or tuple.
        See _format_inputs for details on parsing inputs.

    Returns
    ------
    An Audio object
    """
    # First we organize our inputs.
    inputs = _format_inputs(inputs)

    max_time = 0.0
    sample_rate = 44100
    array_length = 20 * 60  # 20 minutes!
    array_shape = (2, sample_rate * array_length)
    sparse_array = lil_matrix(array_shape)

    initial_offset = 0
    for i, (time_slice, start_time) in enumerate(inputs):
        # if we have a mono file, we return stereo here.
        resampled_audio, left_offset, right_offset = time_slice.get_samples()

        # set the initial offset, so we don't miss the start of the array
        if i == 0:
            initial_offset = max(left_offset * -1, right_offset * -1)

        # get the target start and duration
        start_time = start_time.delta * 1e-9
        duration = time_slice.duration.delta * 1e-9

        # find the max time
        if start_time + duration > max_time:
            max_time = start_time + duration
        # error if we'd go too far
        if start_time + duration > array_length:
            raise SynthesizeError("Amen can only synthesize up to 20 minutes of audio.")

        # get the target start and end samples
        starting_sample, _ = librosa.time_to_samples(
            [start_time, start_time + duration], sr=time_slice.audio.sample_rate
        )

        # figure out the actual starting and ending samples for each channel
        left_start = starting_sample + left_offset + initial_offset
        right_start = starting_sample + right_offset + initial_offset

        # add the data from each channel to the array
        sparse_array[
            0, left_start : left_start + len(resampled_audio[0])
        ] += resampled_audio[0]
        sparse_array[
            1, right_start : right_start + len(resampled_audio[1])
        ] += resampled_audio[1]

    max_samples = librosa.time_to_samples([max_time], sr=sample_rate)
    truncated_array = sparse_array[:, 0 : max_samples[0]].toarray()

    return Audio(raw_samples=truncated_array, sample_rate=sample_rate)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''AMEN Exception classes'''


class AmenError(Exception):
    """
    The root amen exception class
    """

    pass


class SynthesizeError(AmenError):
    """
    Exception class for errors in synthesize.py
    """

    pass


class FeatureError(AmenError):
    """
    Exception class for errors in feature.py
    """

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Utilities'''

import pkg_resources


def example_audio_file():
    """Get the included example file"""
    path = 'example_audio/amen.wav'
    return pkg_resources.resource_filename(__name__, path)


def example_mono_audio_file():
    """Get the included example file"""
    path = 'example_audio/amen-mono.wav'
    return pkg_resources.resource_filename(__name__, path)



import json
import datetime
import collections

import amen.audio
import amen.feature


class AudioAnalysis(object):
    """
    An Analysis object compatible with EchoNest Remix
    """

    def __init__(self, audio_object):
        """
        AudioAnalysis constructor.
        Create an EchoNest compatible Analysis object from an Audio object

        Parameters
        ----------
        audio_object: Audio object

        Returns
        ------
        An AudioAnalysis object
        """
        self.audio = audio_object
        self.sections = AudioQuantumList(kind='sections')
        self.bars = AudioQuantumList(kind='bars')
        self.beats = AudioQuantumList(kind='beats')
        self.tatums = AudioQuantumList(kind='tatums')
        self.segments = AudioQuantumList(kind='segments')

        for key in self.audio.timings:
            # Track-level attributes are dealt with below
            if key == 'track':
                continue
            else:
                setattr(self, key, self._get_quantums(key))

        # TODO: stub segments to be same as beats for now
        self.segments = self._get_quantums('beats')
        self.segments.kind = 'segments'
        for quantum in self.segments:
            quantum.kind = 'segment'

        # Set track-level attributes
        track_level = self.audio.timings['track']
        self.tempo = self.audio.features['tempo'].at(track_level)

    def _get_quantums(self, kind):
        """
        Format Audio Beats to be compatible with EchoNest Remix

        Parameters
        ----------
        kind: String, one of {beats, bars, tatums, sections, segments}

        Returns
        ------
        An AudioQuantumList
        """
        q_list = AudioQuantumList(kind=kind)
        time_slices = self.audio.timings[kind]
        for time_slice in time_slices:
            features = {}
            for key, feature in self.audio.features.items():
                feature = feature.at(time_slice)
                features[key] = feature
            quantum = AudioQuantum(
                kind[:-1], time_slice.time, time_slice.duration, features
            )
            q_list.append(quantum)
        return q_list

    def as_serializable(self):
        """
        Return this object as a serializable dict that contains only dicts,
        lists, and primitives.

        This may be useful, for instance, if you would like to use
        flask.jsonify(analysis.as_serializable())
        instead of the built-in to_json()
        """

        def as_dict(obj):
            # Convert timedeltas to float
            if isinstance(obj, datetime.timedelta):
                return obj.total_seconds()
            # Convert Audio objects to string
            if isinstance(obj, amen.audio.Audio):
                return obj.file_path
            # Convert Feature objects to list
            if isinstance(obj, amen.feature.Feature):
                return as_dict(obj.data.values.tolist())
            if isinstance(obj, collections.Sequence):
                if isinstance(obj, (str, bytes)):
                    return obj
                items = []
                for item in obj:
                    items.append(as_dict(item))
                # Extract if list contains only 1 item
                while isinstance(items, collections.Sequence) and len(items) == 1:
                    items = items[0]
                return items
            if hasattr(obj, '__dict__'):
                result = {}
                for key, value in obj.__dict__.items():
                    if key.startswith('_'):
                        continue
                    result[key] = as_dict(value)
                return result
            # default
            return obj

        return as_dict(self)

    def to_json(self):
        """
        Return this object as a JSON encoded string.
        """
        return json.dumps(self.as_serializable())


class AudioQuantum(object):
    def __init__(self, kind, start, duration, features, confidence=None):
        self.kind = kind
        self.start = start
        self.duration = duration
        self.confidence = confidence
        # Map Amen feature names to Remix names
        for feature, value in features.items():
            key = feature
            if feature == 'chroma':
                # pitches are a list in Remix
                key = 'pitches'
                pitch_names = [
                    'c',
                    'c#',
                    'd',
                    'eb',
                    'e',
                    'f',
                    'f#',
                    'g',
                    'ab',
                    'a',
                    'bb',
                    'b',
                ]
                value = [value[pitch] for pitch in pitch_names]
            elif feature == 'timbre':
                # timbre is a list of 12 PCA things in Remix
                # we take the first 12 MFCCs
                timbre_dimensions = range(12)
                value = [value["mfcc_%s" % t] for t in timbre_dimensions]

            setattr(self, key, value)
        # TODO: stub fake loudness_max values from RMS energy
        self.loudness_max = features.get('amplitude')
        self.loudness_max_time = start
        self.loudness_start = start


class AudioQuantumList(list):
    def __init__(self, initial=[], kind=None):
        list.__init__(self)
        self.kind = kind
        self.extend(initial)



from setuptools import setup, find_packages

import imp

version = imp.load_source('amen.version', 'amen/version.py')
setup(
    name='amen',
    version=version.version,
    description='Algorithmic music remixing',
    url='http://github.com/algorithmic-music-exploration/amen',
    download_url='http://github.com/algorithmic-music-exploration/amen/releases',
    packages=find_packages(),
    package_data={'amen': ['example_audio/*.wav']},
    classifiers=[
        "License :: OSI Approved :: ISC License (ISCL)",
        "Programming Language :: Python",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords='audio music sound',
    license='ISC',
    install_requires=[
        'librosa >= 0.7.0',
        'pandas >= 0.16.0',
        'pysoundfile >= 0.8',
        'six >= 1.10.0',
    ],
)
